/*
 * conf.c - configuration parser for Modbus/TCP server
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <dotconf.h>
#include <limits.h>
#include <pwd.h>
#include <grp.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "log.h"
#include "xmalloc.h"
#include "misc.h"

#include "conf.h"

static const char *opt_port(command_t *cmd, context_t *opaque);
static const char *opt_sockfile(command_t *cmd, context_t *opaque);
static const char *opt_input_image(command_t *cmd, context_t *opaque);
static const char *opt_output_image(command_t *cmd, context_t *opaque);
static const char *opt_max_sessions(command_t *cmd, context_t *opaque);
static const char *opt_keep_alive(command_t *cmd, context_t *opaque);
static const char *opt_inputs(command_t *cmd, context_t *opaque);
static const char *opt_inputs_offset(command_t *cmd, context_t *opaque);
static const char *opt_coils(command_t *cmd, context_t *opaque);
static const char *opt_coils_offset(command_t *cmd, context_t *opaque);
static const char *opt_input_regs(command_t *cmd, context_t *opaque);
static const char *opt_input_regs_offset(command_t *cmd, context_t *opaque);
static const char *opt_holding_regs(command_t *cmd, context_t *opaque);
static const char *opt_holding_regs_offset(command_t *cmd, context_t *opaque);
static const char *opt_loglevel(command_t *cmd, context_t *opaque);
static const char *opt_group(command_t *cmd, context_t *opaque);
static const char *opt_user(command_t *cmd, context_t *opaque);

static configoption_t options[] = {
	{ "port", ARG_INT, opt_port, NULL, 0},
	{ "sockfile", ARG_STR, opt_sockfile, NULL, 0},
	{ "loglevel", ARG_STR, opt_loglevel, NULL, 0},
	{ "group", ARG_STR, opt_group, NULL, 0},
	{ "user", ARG_STR, opt_user, NULL, 0},
	{ "input_image", ARG_INT, opt_input_image, NULL, 0},
	{ "output_image", ARG_INT, opt_output_image, NULL, 0},
	{ "max_sessions", ARG_INT, opt_max_sessions, NULL, 0},
	{ "keep_alive", ARG_INT, opt_keep_alive, NULL, 0},
	{ "inputs", ARG_INT, opt_inputs, NULL, 0},
	{ "inputs_offset", ARG_INT, opt_inputs_offset, NULL, 0},
	{ "coils", ARG_INT, opt_coils, NULL, 0},
	{ "coils_offset", ARG_INT, opt_coils_offset, NULL, 0},
	{ "input_regs", ARG_INT, opt_input_regs, NULL, 0},
	{ "input_regs_offset", ARG_INT, opt_input_regs_offset, NULL, 0},
	{ "holding_regs", ARG_INT,  opt_holding_regs, NULL, 0},
	{ "holding_regs_offset", ARG_INT, opt_holding_regs_offset, NULL, 0},
	LAST_OPTION,
};

static const char *opt_port(command_t *cmd, context_t *opaque)
{
	struct modbusd_options *opt = (struct modbusd_options *) opaque;
	opt->port = cmd->data.value;
	return NULL;
}

static const char *opt_sockfile(command_t *cmd, context_t *opaque)
{
	struct modbusd_options *opt = (struct modbusd_options *) opaque;

	if (cmd->data.str[0] == '/')
		xstrncpy(opt->sockfile, cmd->data.str, sizeof(opt->sockfile));
	else
		snprintf(opt->sockfile, sizeof(opt->sockfile), "%s/%s/%s",
				CONFIG_DIR, DAEMON_DIR, cmd->data.str);
	return NULL;
}

static const char *opt_input_image(command_t *cmd, context_t *opaque)
{
	struct modbusd_options *opt = (struct modbusd_options *) opaque;
	opt->input_image = cmd->data.value;
	return NULL;
}

static const char *opt_output_image(command_t *cmd, context_t *opaque)
{
	struct modbusd_options *opt = (struct modbusd_options *) opaque;
	opt->output_image = cmd->data.value;
	return NULL;
}

static const char *opt_max_sessions(command_t *cmd, context_t *opaque)
{
	struct modbusd_options *opt = (struct modbusd_options *) opaque;

	if (cmd->data.value == 0)
		opt->max_sessions = 0;
	else if (cmd->data.value < MIN_SESSIONS ||
	   cmd->data.value > MAX_SESSIONS)
		dotconf_warning(cmd->configfile, DCLOG_WARNING, 0,
				"option 'max_sessions' argument out of range %d..%d",
				MIN_SESSIONS, MAX_SESSIONS);
	else
		opt->max_sessions = cmd->data.value;
	return NULL;
}

static const char *opt_keep_alive(command_t *cmd, context_t *opaque)
{
	struct modbusd_options *opt = (struct modbusd_options *) opaque;

	if (cmd->data.value == 0)
		opt->keep_alive = 0;
	else if (cmd->data.value < MIN_KEEP_ALIVE ||
	   cmd->data.value > MAX_KEEP_ALIVE)
		dotconf_warning(cmd->configfile, DCLOG_WARNING, 0,
				"option 'keep_alive' argument out of range %d..%d",
				MIN_KEEP_ALIVE, MAX_KEEP_ALIVE);
	else
		opt->keep_alive = cmd->data.value * 1000;
	return NULL;
}

static const char *opt_inputs(command_t *cmd, context_t *opaque)
{
	struct modbusd_options *opt = (struct modbusd_options *) opaque;
	opt->inputs = cmd->data.value;
	return NULL;
}

static const char *opt_inputs_offset(command_t *cmd, context_t *opaque)
{
	struct modbusd_options *opt = (struct modbusd_options *) opaque;
	opt->inputs_offset = cmd->data.value;
	return NULL;
}

static const char *opt_coils(command_t *cmd, context_t *opaque)
{
	struct modbusd_options *opt = (struct modbusd_options *) opaque;
	opt->coils = cmd->data.value;
	return NULL;
}

static const char *opt_coils_offset(command_t *cmd, context_t *opaque)
{
	struct modbusd_options *opt = (struct modbusd_options *) opaque;
	opt->coils_offset = cmd->data.value;
	return NULL;
}

static const char *opt_input_regs(command_t *cmd, context_t *opaque)
{
	struct modbusd_options *opt = (struct modbusd_options *) opaque;

	if (cmd->data.value % 2) {
		dotconf_warning(cmd->configfile, DCLOG_ERR, 0,
				"option 'input_regs' must be multiple of two");
	} else {
		opt->input_regs = cmd->data.value;
	}
	return NULL;
}

static const char *opt_input_regs_offset(command_t *cmd, context_t *opaque)
{
	struct modbusd_options *opt = (struct modbusd_options *) opaque;

	if (cmd->data.value % 2) {
		dotconf_warning(cmd->configfile, DCLOG_ERR, 0,
				"option 'input_regs_offset' must be multiple of two");
	} else {
		opt->input_regs_offset = cmd->data.value;
	}
	return NULL;
}

static const char *opt_holding_regs(command_t *cmd, context_t *opaque)
{
	struct modbusd_options *opt = (struct modbusd_options *) opaque;

	if (cmd->data.value % 2) {
		dotconf_warning(cmd->configfile, DCLOG_ERR, 0,
				"option 'holding_regs' must be miltiple of two");
	} else {
		opt->holding_regs = cmd->data.value;
	}
	return NULL;
}

static const char *opt_holding_regs_offset(command_t *cmd, context_t *opaque)
{
	struct modbusd_options *opt = (struct modbusd_options *) opaque;

	if (cmd->data.value % 2) {
		dotconf_warning(cmd->configfile, DCLOG_ERR, 0,
				"option 'holding_regs_offset' must be multiple of two");
	} else {
		opt->holding_regs_offset = cmd->data.value;
	}
	return NULL;
}

static const char *opt_loglevel(command_t *cmd, context_t *opaque)
{
	char *ptr, *c;
	struct modbusd_options *opt = (struct modbusd_options *) opaque;

	c = cmd->data.str;
	ptr = strchr(c, '.');
	if (ptr) {
		*ptr++ = 0;
		if (strcmp(c, "daemon") == 0)
			opt->facility = LOG_FACILITY_DAEMON;
		else if (strcmp(c, "local0") == 0)
			opt->facility = LOG_FACILITY_LOCAL0;
		else if (strcmp(c, "local1") == 0)
			opt->facility = LOG_FACILITY_LOCAL1;
		else if (strcmp(c, "local2") == 0)
			opt->facility = LOG_FACILITY_LOCAL2;
		else if (strcmp(c, "local3") == 0)
			opt->facility = LOG_FACILITY_LOCAL3;
		else if (strcmp(c, "local4") == 0)
			opt->facility = LOG_FACILITY_LOCAL4;
		else if (strcmp(c, "local5") == 0)
			opt->facility = LOG_FACILITY_LOCAL5;
		else if (strcmp(c, "local6") == 0)
			opt->facility = LOG_FACILITY_LOCAL6;
		else if (strcmp(c, "local7") == 0)
			opt->facility = LOG_FACILITY_LOCAL7;
		else if (strcmp(c, "syslog") == 0)
			opt->facility = LOG_FACILITY_SYSLOG;
		else if (strcmp(c, "user") == 0)
			opt->facility = LOG_FACILITY_USER;
		else {
			dotconf_warning(cmd->configfile, DCLOG_WARNING, 0,
					"option 'loglevel' has unknown facility '%.100s'",
					ptr);
			return NULL;
		}
		opt->enable_syslog = 1;
		c = ptr;
	}
	if (strcmp(c, "debug") == 0)
		opt->level = LOG_LEVEL_DEBUG;
	else if (strcmp(c, "verbose") == 0)
		opt->level = LOG_LEVEL_VERBOSE;
	else if (strcmp(c, "info") == 0)
		opt->level = LOG_LEVEL_INFO;
	else if (strcmp(c, "warn") == 0)
		opt->level = LOG_LEVEL_WARN;
	else if (strcmp(c, "error") == 0)
		opt->level = LOG_LEVEL_ERROR;
	else {
		dotconf_warning(cmd->configfile, DCLOG_WARNING, 0,
				"option 'loglevel' has unknown value '%.100s'",
				c);
		return NULL;
	}
	return NULL;
}

static const char *opt_user(command_t *cmd, context_t *opaque)
{
	int res;
	struct passwd pwd;
	struct modbusd_options *opt = (struct modbusd_options *) opaque;

	res = getpw_by_name_or_uid(cmd->data.str, &pwd);
	if (res) {
		dotconf_warning(cmd->configfile, DCLOG_WARNING, 0,
				"option 'user' gives nonexisting uid or username");
		return NULL;
	}
	opt->user = pwd.pw_uid;
	if (opt->group == 0)
		opt->group = pwd.pw_gid;
	return NULL;
}

static const char *opt_group(command_t *cmd, context_t *opaque)
{
	int res;
	struct group grp;
	struct modbusd_options *opt = (struct modbusd_options *) opaque;

	res = getgr_by_name_or_gid(cmd->data.str, &grp);
	if (res) {
		dotconf_warning(cmd->configfile, DCLOG_WARNING, 0,
				"option 'group' gives nonexisting gid or group");
		return NULL;
	}
	opt->group = grp.gr_gid;
	return NULL;
}

static const int error_handler(configfile_t *config, int type,
		long dc_errno, const char *msg)
{
	warn("%s: file '%s' line %ld", msg, config->filename, config->line);
	return 0;
}

int parse_config(char *configfile, struct modbusd_options *opt)
{
	const char *err;
	configfile_t *config;

	if (!opt)
		return -1;

	opt->port = MODBUS_PORT;
	opt->max_sessions = DEFAULT_MAX_SESSIONS;
	opt->keep_alive = DEFAULT_KEEP_ALIVE;
	snprintf(opt->sockfile, sizeof(opt->sockfile), "%s/%s/%s",
			LOCALSTATE_DIR, DAEMON_DIR, DEFAULT_SOCKFILE);
	opt->level = LOG_LEVEL_ERROR;
	opt->facility = LOG_FACILITY_DAEMON;
	opt->user = 0;
	opt->group = 0;
	opt->input_image = 0;
	opt->output_image = 0;
	opt->coils = 0;
	opt->coils_offset = 0;
	opt->inputs = 0;
	opt->inputs_offset = 0;
	opt->input_regs = 0;
	opt->input_regs_offset = 0;
	opt->holding_regs = 0;
	opt->holding_regs_offset = 0;
	config = dotconf_create(configfile, options, (context_t *) opt, DONT_SUBSTITUTE);
	if (!config) {
		verbose("can't read configuration file '%s'", configfile);
		return -1;
	}
	config->errorhandler = (dotconf_errorhandler_t) error_handler;
	config->contextchecker = NULL;
	while (!config->eof) {
		err = dotconf_command_loop_until_error(config);
		if (err) {
			dotconf_warning(config, DCLOG_WARNING, 0, "%s", err);
		} else
			break;
	}
	dotconf_cleanup(config);
	if (!opt->input_image && !opt->output_image) {
		warn("modbus has no inputs and no outputs image");
		return -1;
	}
	if (opt->input_image) {
		if (opt->inputs_offset + opt->inputs > opt->input_image) {
			warn("modbus inputs extend beyound input image");
			return -1;
		}
		if (opt->input_regs + opt->input_regs_offset > opt->input_image) {
			warn("modbus input registers extend beyound input image");
			return -1;
		}
	} else if (opt->inputs || opt->inputs_offset ||
	     opt->input_regs || opt->input_regs_offset) {
		warn("input image has zero size but "
				"inputs/regs present in configuration");
		return -1;
	}
	if (opt->output_image) {
		if (opt->coils_offset + opt->coils > opt->output_image) {
			warn("modbus coils extend beyound output image");
			return -1;
		}
		if (opt->holding_regs_offset + opt->holding_regs > opt->output_image) {
			warn("modbus holding registers extend beyound output image");
			return -1;
		}
	} else if (opt->coils || opt->coils_offset ||
	     opt->holding_regs || opt->holding_regs_offset) {
		warn("output image has zero size but "
				"coils/holding regs present in configuration");
		return -1;
	}
	return 0;
}
